function mi = kraskovmi(x1,x2,k)
% Compute the mutual information (in bits) using the method of Kraskov, Stoegbauer and Grassberger
% "Estimating Mutual Information", arXiv:cond-mat/0305641 v1 28/5/03.
%
% kraskovmi(x1,x2,k)
%
% Arguments x1 and x2 are row vectors
%
% Argument k is the neighbor's order in k-nearset-neighbor
%
% Doesn't handle strictly equal distances between data points well. Due to this we add a small amount
% of noise to the data.

if size(x1,1) ~= 1 | size(x2) ~= size(x1)
   fprintf(1,'Wrong argument sizes\n')
   return
end

npoints = size(x1,2);

% Add small noise to avoid problems with repeated (e.g. quantized) values

x1 = x1 + (rand(size(x1)) - .5) / (1e6 * max(abs(x1)));
x2 = x2 + (rand(size(x2)) - .5) / (1e6 * max(abs(x2)));

% Sort x1 (and x2 accordingly) for a quicker search

[x1,x1index] = sort(x1);
x2 = x2(x1index);

for i = 1:npoints
   
   distx1 = abs(x1 - x1(i));
   distx2 = abs(x2 - x2(i));
   [distsorted,distindex] = sort(max(distx1,distx2));
   neighbdist(i) = distsorted(k+1);
   neighbindex(i) = distindex(k+1);
   
   n1(i) = sum(distx1 < neighbdist(i)) - 1;    % -1 to discount point i itself
   n2(i) = sum(distx2 < neighbdist(i)) - 1;    % -1 to discount point i itself
   
end

% Compute the MI estimate

mi = psi(k) - (sum(psi(n1+1)) + sum(psi(n2+1))) / npoints + psi(npoints);

mi = mi / log(2);    % convert to bits units
